using System;
namespace symmetry
{
	
	/// <summary> A quad-tree based implementation of the Barnes-Hut algorithm for
	/// efficient n-body force simulations. This can be used to create
	/// force-based layouts of graphs. All input NodeItems to the simulation
	/// are expected to have a "mass" attribute of type float, retrievable 
	/// by the call item.getVizAttribute("mass").
	/// 
	/// The Barnes-Hut algorithm originally appeared in <i>"A Hierarchical 
	/// O(n log n)force calculation algorithm" by J. Barnes and P. Hut, Nature, 
	/// v.324, December 1986</i>. For more details on the algorithm, see one of
	/// the following links --
	/// <ul>
	/// <li><a href="http://www.cs.berkeley.edu/~demmel/cs267/lecture26/lecture26.html">James Demmel's UC Berkeley lecture notes</a>
	/// <li><a href="http://www.physics.gmu.edu/~large/lr_forces/desc/bh/bhdesc.html">Description of the Barnes-Hut algorithm</a>
	/// <li><a href="http://www.ifa.hawaii.edu/~barnes/treecode/treeguide.html">Joshua Barnes' recent implementation</a>
	/// </ul>
	/// 
	/// </summary>
	/// <version>  1.0
	/// </version>
	/// <author>  <a href="http://jheer.org">Jeffrey Heer</a> prefuse(AT)jheer.org
	/// </author>
	public class NBodyForce:AbstractForce
	{
		private void  InitBlock()
		{
			factory = new QuadTreeNodeFactory(this);
		}
		override public bool ItemForce
		{
			get
			{
				return true;
			}
			//
			
		}
		override protected internal System.String[] ParameterNames
		{
			get
			{
				return pnames;
			}
			//
			
		}
		
		/* 
		* The labeling scheme for quadtree child nodes goes row by row.
		*   0 | 1    0 -> top left,    1 -> top right
		*  -------
		*   2 | 3    2 -> bottom left, 3 -> bottom right
		*/
		
		private static System.String[] pnames = new System.String[]{"GravitationalConstant", "MinimumDistance", "BarnesHutTheta"};
		
		public const float DEFAULT_GRAV_CONSTANT = - 0.4f;
		public const float DEFAULT_MIN_DISTANCE = - 1f;
		public const float DEFAULT_THETA = 0.9f;
		public const int GRAVITATIONAL_CONST = 0;
		public const int MIN_DISTANCE = 1;
		public const int BARNES_HUT_THETA = 2;
		
		private float xMin, xMax, yMin, yMax;
		//UPGRADE_NOTE: The initialization of  'factory' was moved to method 'InitBlock'. 'ms-help://MS.VSCC.2003/commoner/redir/redirect.htm?keyword="jlca1005"'
		private QuadTreeNodeFactory factory;
		private QuadTreeNode root;
		
		/// <summary> Construct a new empty NBodyForce simulation.</summary>
		public NBodyForce():this(DEFAULT_GRAV_CONSTANT, DEFAULT_MIN_DISTANCE, DEFAULT_THETA)
		{
		} //
		
		public NBodyForce(float gravConstant, float minDistance, float theta)
		{
			InitBlock();
			params_Renamed = new float[]{gravConstant, minDistance, theta};
			root = factory.QuadTreeNode;
		} //
		
		/// <summary> Set the bounds of the region for which to compute the n-body simulation</summary>
		/// <param name="xMin">the minimum x-coordinate
		/// </param>
		/// <param name="yMin">the minimum y-coordinate
		/// </param>
		/// <param name="xMax">the maximum x-coordinate
		/// </param>
		/// <param name="yMax">the maximum y-coordinate
		/// </param>
		private void  setBounds(int xMin, int yMin, int xMax, int yMax)
		{
			this.xMin = xMin;
			this.yMin = yMin;
			this.xMax = xMax;
			this.yMax = yMax;
		} //
		
		/// <summary> Clears the quadtree of all entries.</summary>
		public virtual void  clear()
		{
			clearHelper(root);
			root = factory.QuadTreeNode;
		} //
		
		private void  clearHelper(QuadTreeNode n)
		{
			for (int i = 0; i < n.children.Length; i++)
			{
				if (n.children[i] != null)
					clearHelper(n.children[i]);
			}
			factory.reclaim(n);
		} //
		
		/// <summary> Initialize the simulation with the provided enclosing simulation. After
		/// this call has been made, the simulation can be queried for the 
		/// n-body force on a given item.
		/// </summary>
		/// <param name="fsim">the enclosing ForceSimulator
		/// </param>
		public override void  init(ForceSimulator fsim)
		{
			clear(); // clear internal state
			
			// compute and squarify bounds of quadtree
			int x1 = System.Int32.MaxValue, y1 = System.Int32.MaxValue;
			int x2 = System.Int32.MinValue, y2 = System.Int32.MinValue;
			System.Collections.IEnumerator itemIter = fsim.Items;
			//UPGRADE_TODO: Method 'java.util.Iterator.hasNext' was converted to 'System.Collections.IEnumerator.MoveNext' which has a different behavior. 'ms-help://MS.VSCC.2003/commoner/redir/redirect.htm?keyword="jlca1073_javautilIteratorhasNext"'
			while (itemIter.MoveNext())
			{
				//UPGRADE_TODO: Method 'java.util.Iterator.next' was converted to 'System.Collections.IEnumerator.Current' which has a different behavior. 'ms-help://MS.VSCC.2003/commoner/redir/redirect.htm?keyword="jlca1073_javautilIteratornext"'
				ForceItem item = (ForceItem) itemIter.Current;
				//UPGRADE_TODO: Method 'java.lang.Math.round' was converted to 'System.Math.Round' which has a different behavior. 'ms-help://MS.VSCC.2003/commoner/redir/redirect.htm?keyword="jlca1073"'
				int x = (int) System.Math.Round((double) item.location[0]);
				//UPGRADE_TODO: Method 'java.lang.Math.round' was converted to 'System.Math.Round' which has a different behavior. 'ms-help://MS.VSCC.2003/commoner/redir/redirect.htm?keyword="jlca1073"'
				int y = (int) System.Math.Round((double) item.location[1]);
				if (x < x1)
					x1 = x;
				if (y < y1)
					y1 = y;
				if (x > x2)
					x2 = x;
				if (y > y2)
					y2 = y;
			}
			int dx = x2 - x1, dy = y2 - y1;
			if (dx > dy)
			{
				y2 = y1 + dx;
			}
			else
			{
				x2 = x1 + dy;
			}
			setBounds(x1, y1, x2, y2);
			
			// insert items into quadtree
			itemIter = fsim.Items;
			//UPGRADE_TODO: Method 'java.util.Iterator.hasNext' was converted to 'System.Collections.IEnumerator.MoveNext' which has a different behavior. 'ms-help://MS.VSCC.2003/commoner/redir/redirect.htm?keyword="jlca1073_javautilIteratorhasNext"'
			while (itemIter.MoveNext())
			{
				//UPGRADE_TODO: Method 'java.util.Iterator.next' was converted to 'System.Collections.IEnumerator.Current' which has a different behavior. 'ms-help://MS.VSCC.2003/commoner/redir/redirect.htm?keyword="jlca1073_javautilIteratornext"'
				ForceItem item = (ForceItem) itemIter.Current;
				insert(item);
			}
			
			// calculate magnitudes and centers of mass
			calcMass(root);
		} //
		
		/// <summary> Inserts an item into the quadtree.</summary>
		/// <param name="item">the ForceItem to add.
		/// @throws IllegalStateException if the current location of the item is
		/// outside the bounds of the quadtree
		/// </param>
		public virtual void  insert(ForceItem item)
		{
			// insert item into the quadtrees
			insert(item, root, xMin, yMin, xMax, yMax);
		} //
		
		private void  insert(ForceItem p, QuadTreeNode n, float x1, float y1, float x2, float y2)
		{
			// try to insert particle p at node n in the quadtree
			// by construction, each leaf will contain either 1 or 0 particles
			if (n.hasChildren)
			{
				// n contains more than 1 particle
				insertHelper(p, n, x1, y1, x2, y2);
			}
			else if (n.value_Renamed != null)
			{
				// n contains 1 particle
				if (isSameLocation(n.value_Renamed, p))
				{
					insertHelper(p, n, x1, y1, x2, y2);
				}
				else
				{
					ForceItem v = n.value_Renamed; n.value_Renamed = null;
					insertHelper(v, n, x1, y1, x2, y2);
					insertHelper(p, n, x1, y1, x2, y2);
				}
			}
			else
			{
				// n is empty, so is a leaf
				n.value_Renamed = p;
			}
		} //
		
		private static bool isSameLocation(ForceItem f1, ForceItem f2)
		{
			float dx = System.Math.Abs(f1.location[0] - f2.location[0]);
			float dy = System.Math.Abs(f1.location[1] - f2.location[1]);
			return (dx < 0.01 && dy < 0.01);
		} //
		
		private void  insertHelper(ForceItem p, QuadTreeNode n, float x1, float y1, float x2, float y2)
		{
			float x = p.location[0], y = p.location[1];
			float splitx = x1 + ((x2 - x1) / 2);
			float splity = y1 + ((y2 - y1) / 2);
			int i = (x >= splitx?1:0) + (y >= splity?2:0);
			// create new child node, if necessary
			if (n.children[i] == null)
			{
				n.children[i] = factory.QuadTreeNode;
				n.hasChildren = true;
			}
			// update bounds
			if (i == 1 || i == 3)
				x1 = splitx;
			else
				x2 = splitx;
			if (i > 1)
				y1 = splity;
			else
				y2 = splity;
			// recurse 
			insert(p, n.children[i], x1, y1, x2, y2);
		} //
		
		private void  calcMass(QuadTreeNode n)
		{
			float xcom = 0, ycom = 0;
			n.mass = 0;
			if (n.hasChildren)
			{
				for (int i = 0; i < n.children.Length; i++)
				{
					if (n.children[i] != null)
					{
						calcMass(n.children[i]);
						n.mass += n.children[i].mass;
						xcom += n.children[i].mass * n.children[i].com[0];
						ycom += n.children[i].mass * n.children[i].com[1];
					}
				}
			}
			if (n.value_Renamed != null)
			{
				n.mass += n.value_Renamed.mass;
				xcom += n.value_Renamed.mass * n.value_Renamed.location[0];
				ycom += n.value_Renamed.mass * n.value_Renamed.location[1];
			}
			n.com[0] = xcom / n.mass;
			n.com[1] = ycom / n.mass;
		} //
		
		/// <summary> Calculates the force vector acting on the given item.</summary>
		/// <param name="item">the ForceItem for which to compute the force
		/// </param>
		public override void  getForce(ForceItem item)
		{
			forceHelper(item, root, xMin, yMin, xMax, yMax);
		} //
		
		private void  forceHelper(ForceItem item, QuadTreeNode n, float x1, float y1, float x2, float y2)
		{
			float dx = n.com[0] - item.location[0];
			float dy = n.com[1] - item.location[1];
			//UPGRADE_WARNING: Narrowing conversions may produce unexpected results in C#. 'ms-help://MS.VSCC.2003/commoner/redir/redirect.htm?keyword="jlca1042"'
			float r = (float) System.Math.Sqrt(dx * dx + dy * dy);
			bool same = false;
			if (r == 0.0)
			{
				// if items are in the exact same place, add some noise
				//UPGRADE_WARNING: Narrowing conversions may produce unexpected results in C#. 'ms-help://MS.VSCC.2003/commoner/redir/redirect.htm?keyword="jlca1042"'
				dx = ((float) SupportClass.Random.NextDouble() - 0.5f) / 50.0f;
				//UPGRADE_WARNING: Narrowing conversions may produce unexpected results in C#. 'ms-help://MS.VSCC.2003/commoner/redir/redirect.htm?keyword="jlca1042"'
				dy = ((float) SupportClass.Random.NextDouble() - 0.5f) / 50.0f;
				//UPGRADE_WARNING: Narrowing conversions may produce unexpected results in C#. 'ms-help://MS.VSCC.2003/commoner/redir/redirect.htm?keyword="jlca1042"'
				r = (float) System.Math.Sqrt(dx * dx + dy * dy);
				same = true;
			}
			bool minDist = params_Renamed[MIN_DISTANCE] > 0f && r > params_Renamed[MIN_DISTANCE];
			
			// the Barnes-Hut approximation criteria is if the ratio of the
			// size of the quadtree box to the distance between the point and
			// the box's center of mass is beneath some threshold theta.
			if ((!n.hasChildren && n.value_Renamed != item) || (!same && (x2 - x1) / r < params_Renamed[BARNES_HUT_THETA]))
			{
				if (minDist)
					return ;
				// either only 1 particle or we meet criteria
				// for Barnes-Hut approximation, so calc force
				float v = params_Renamed[GRAVITATIONAL_CONST] * item.mass * n.mass / (r * r * r);
				item.force[0] += v * dx;
				item.force[1] += v * dy;
			}
			else if (n.hasChildren)
			{
				// recurse for more accurate calculation
				float splitx = x1 + ((x2 - x1) / 2);
				float splity = y1 + ((y2 - y1) / 2);
				for (int i = 0; i < n.children.Length; i++)
				{
					if (n.children[i] != null)
					{
						forceHelper(item, n.children[i], (i == 1 || i == 3?splitx:x1), (i > 1?splity:y1), (i == 1 || i == 3?x2:splitx), (i > 1?y2:splity));
					}
				}
				if (minDist)
					return ;
				if (n.value_Renamed != null && n.value_Renamed != item)
				{
					float v = params_Renamed[GRAVITATIONAL_CONST] * item.mass * n.value_Renamed.mass / (r * r * r);
					item.force[0] += v * dx;
					item.force[1] += v * dy;
				}
			}
		} //
		
		//UPGRADE_NOTE: Field 'EnclosingInstance' was added to class 'QuadTreeNode' to access its enclosing instance. 'ms-help://MS.VSCC.2003/commoner/redir/redirect.htm?keyword="jlca1019"'
		/// <summary> Represents a node in the quadtree.</summary>
		public sealed class QuadTreeNode
		{
			private void  InitBlock(NBodyForce enclosingInstance)
			{
				this.enclosingInstance = enclosingInstance;
			}
			private NBodyForce enclosingInstance;
			public NBodyForce Enclosing_Instance
			{
				get
				{
					return enclosingInstance;
				}
				
			}
			public QuadTreeNode(NBodyForce enclosingInstance)
			{
				InitBlock(enclosingInstance);
				com = new float[]{0.0f, 0.0f};
				children = new QuadTreeNode[4];
			} //
			internal bool hasChildren = false;
			internal float mass; // total mass held by this node
			internal float[] com; // center of mass of this node 
			internal ForceItem value_Renamed; // ForceItem in this node, null if node has children
			internal QuadTreeNode[] children; // children nodes
		} // end of inner class QuadTreeNode
		
		//UPGRADE_NOTE: Field 'EnclosingInstance' was added to class 'QuadTreeNodeFactory' to access its enclosing instance. 'ms-help://MS.VSCC.2003/commoner/redir/redirect.htm?keyword="jlca1019"'
		/// <summary> Helper class to minimize number of object creations across multiple
		/// uses of the quadtree.
		/// </summary>
		public sealed class QuadTreeNodeFactory
		{
			public QuadTreeNodeFactory(NBodyForce enclosingInstance)
			{
				InitBlock(enclosingInstance);
			}
			private void  InitBlock(NBodyForce enclosingInstance)
			{
				this.enclosingInstance = enclosingInstance;
				//UPGRADE_ISSUE: Class hierarchy differences between 'java.util.ArrayList' and 'System.Collections.ArrayList' may cause compilation errors. 'ms-help://MS.VSCC.2003/commoner/redir/redirect.htm?keyword="jlca1186"'
				nodes = new System.Collections.ArrayList();
			}
			private NBodyForce enclosingInstance;
			public QuadTreeNode QuadTreeNode
			{
				get
				{
					//UPGRADE_TODO: The equivalent in .NET for method 'java.util.ArrayList.size' may return a different value. 'ms-help://MS.VSCC.2003/commoner/redir/redirect.htm?keyword="jlca1043"'
					if (nodes.Count > 0)
					{
						//UPGRADE_TODO: The equivalent in .NET for method 'java.util.ArrayList.remove' may return a different value. 'ms-help://MS.VSCC.2003/commoner/redir/redirect.htm?keyword="jlca1043"'
						//UPGRADE_TODO: The equivalent in .NET for method 'java.util.ArrayList.size' may return a different value. 'ms-help://MS.VSCC.2003/commoner/redir/redirect.htm?keyword="jlca1043"'
						return (QuadTreeNode) nodes.RemoveAt(nodes.Count - 1);
					}
					else
					{
						return new QuadTreeNode();
					}
				}
				//
				
			}
			public NBodyForce Enclosing_Instance
			{
				get
				{
					return enclosingInstance;
				}
				
			}
			private int maxNodes = 10000;
			//UPGRADE_ISSUE: Class hierarchy differences between 'java.util.ArrayList' and 'System.Collections.ArrayList' may cause compilation errors. 'ms-help://MS.VSCC.2003/commoner/redir/redirect.htm?keyword="jlca1186"'
			//UPGRADE_NOTE: The initialization of  'nodes' was moved to method 'InitBlock'. 'ms-help://MS.VSCC.2003/commoner/redir/redirect.htm?keyword="jlca1005"'
			private System.Collections.ArrayList nodes;
			public void  reclaim(QuadTreeNode n)
			{
				n.mass = 0;
				n.com[0] = 0.0f; n.com[1] = 0.0f;
				n.value_Renamed = null;
				n.hasChildren = false;
				SupportClass.ArraysSupport.FillArray(n.children, null);
				//UPGRADE_TODO: The equivalent in .NET for method 'java.util.ArrayList.size' may return a different value. 'ms-help://MS.VSCC.2003/commoner/redir/redirect.htm?keyword="jlca1043"'
				if (nodes.Count < maxNodes)
				{
					//UPGRADE_TODO: The equivalent in .NET for method 'java.util.ArrayList.add' may return a different value. 'ms-help://MS.VSCC.2003/commoner/redir/redirect.htm?keyword="jlca1043"'
					nodes.Add(n);
				}
			} //
		} // end of inner class QuadTreeNodeFactory
	} // end of class NBodyForce
}